/*
**    BPALogin - lightweight portable BIDS2 login client
**    Copyright (c) 2001-3 Shane Hyde, and others.
** 
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation; either version 2 of the License, or
**  (at your option) any later version.
** 
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
** 
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
**
*/ 

/**
 * Changes:
 * 
 * 2001-12-05:  wdrose     Added errno.h to list of UNIX-only includes.
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <netdb.h>
#include <syslog.h>
#include <time.h>
#include <signal.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/sysinfo.h>
#include <sys/ioctl.h>
#include <linux/if.h>
#include <linux/if_ether.h>
#include <linux/if_packet.h>

#define LOGIN_VERSION	1

#define DEFAULT_AUTHSERVER	"login-server"
#define DEFAULT_AUTHDOMAIN	""
#define DEFAULT_AUTHPORT	5050
#define DEFAULT_CONFFILE		"/etc/bpalogin.conf"
#define DEFAULT_DEVNAME		"eth1" /* used for Dial-on-Demand */

/* state engine codes */
#define STATE_NEED_PROTOCOL	0
#define STATE_SEND_PROT_REQ	1
#define STATE_AWAIT_NEG_RESP	2
#define STATE_SEND_LOGIN_REQ	3
#define STATE_AWAIT_LOGIN_AUTH_RESP	4
#define STATE_SEND_LOGIN_AUTH_REQ		5
#define STATE_AWAIT_LOGIN_RESP		6
#define STATE_SEND_LOGOUT_REQ		7
#define STATE_AWAIT_LOGOUT_AUTH_RESP	8
#define STATE_SEND_LOGOUT_AUTH_REQ	9
#define STATE_AWAIT_LOGOUT_RESP	10
#define STATE_IDLE_LOGIN				11
#define STATE_RECEIVED_STATUS_REQ	12
#define STATE_RECEIVED_RESTART_REQ	13
#define STATE_IDLE_LOGOFF			14

/* message type codes */
#define T_MSG_MIN	1
#define T_MSG_PROTOCOL_NEG_REQ	1
#define T_MSG_PROTOCOL_NEG_RESP	2
#define T_MSG_LOGIN_REQ		3
#define T_MSG_LOGIN_AUTH_REQ	4
#define T_MSG_LOGIN_RESP	5
#define T_MSG_LOGOUT_REQ	6
#define T_MSG_LOGOUT_AUTH_RESP	7
#define T_MSG_LOGOUT_RESP		8
#define T_MSG_AUTH_RESP		9
#define T_MSG_AUTH_REQ		10
#define T_MSG_STATUS_REQ	11
#define T_MSG_STATUS_RESP	12
#define T_MSG_RESTART_REQ	13
#define T_MSG_RESTART_RESP	14
#define T_MSG_MAX			14

/* message parameter codes */
#define T_PARAM_MIN		1
#define T_PARAM_PROTOCOL_LIST		1
#define T_PARAM_PROTOCOL_SELECT	2
#define T_PARAM_CLIENT_VERSION		3
#define T_PARAM_OS_IDENTITY			4
#define T_PARAM_OS_VERSION			5
#define T_PARAM_REASON_CODE		6
#define T_PARAM_USERNAME			7
#define T_PARAM_REQ_PORT			8
#define T_PARAM_RESPONSE_TEXT		9
#define T_PARAM_STATUS_CODE		10
#define T_PARAM_AUTH_CREDENTIALS	11
#define T_PARAM_NONCE		12
#define T_PARAM_SEQNUM		13
#define T_PARAM_HASH_METHOD	14
#define T_PARAM_LOGIN_SERVICE_PORT		15
#define T_PARAM_LOGOUT_SERVICE_PORT	16
#define T_PARAM_STATUS_SERVICE_PORT	17
#define T_PARAM_SUSPEND_IND	18
#define T_PARAM_STATUS_AUTH	19
#define T_PARAM_RESTART_AUTH	20
#define T_PARAM_TIMESTAMP	21
#define T_PARAM_TSMLIST		22
#define T_PARAM_LOGIN_PARAM_HASH	23
#define T_PARAM_LOGIN_SERVER_HOST	24
#define T_PARAM_MAX					24

/* login reason codes */
#define T_LOGIN_REASON_CODE_NORMAL	0
#define T_LOGIN_REASON_CODE_REAUTH	1

/* logout reason codes */
#define T_LOGOUT_REASON_CODE_USER_INITIATED	0
#define T_LOGOUT_REASON_CODE_APP_SHUTDOWN	1
#define T_LOGOUT_REASON_CODE_OS_SHUTDOWN	2
#define T_LOGOUT_REASON_CODE_UNKNOWN		3

/* client status transaction codes */
#define T_STATUS_TRANSACTION_OK	0

/* restart reasons */
#define T_RESTART_ADMIN		0

/* auth responses */
#define T_AUTH_NOHASH		0
#define T_AUTH_MD5_HASH	1

/* protocol types */
#define T_PROTOCOL_CHAL		1

/* status return codes */
#define T_STATUS_SUCCESS	0
#define T_STATUS_USERNAME_NOT_FOUND	1
#define T_STATUS_INCORRECT_PASSWORD	2
#define T_STATUS_ACCOUNT_DISABLED		3
#define T_STATUS_USER_DISABLED			4
#define T_STATUS_LOGIN_SUCCESSFUL_ALREADY_LOGGED_IN	100
#define T_STATUS_LOGIN_RETRY_LIMIT						101
#define T_STATUS_LOGIN_SUCCESSFUL_SWVER		102
#define T_STATUS_LOGIN_FAIL_SW			103
#define T_STATUS_LOGOUT_SUCCESSFUL_ALREADY_DISCONNECTED		200
#define T_STATUS_LOGOUT_AUTH_RETRY_LIMIT		201
#define T_STATUS_LOGIN_SUCCESS_SWVER			300
#define T_STATUS_LOGIN_FAIL_SWVER				301
#define T_STATUS_LOGIN_FAIL_INV_PROT			302
#define T_STATUS_LOGIN_UNKNOWN				500
#define T_STATUS_FAIL_USERNAME_VALIDATE		501
#define T_STATUS_FAIL_PASSWORD_VALIDATE		502

/* internal using */
#define T_STATUS_IGNORE_ERR		0xFFFD
#define T_STATUS_ERROR			0xFFFE
#define T_STATUS_TIMEOUT		0xFFFF

typedef char sint8;
typedef unsigned char uint8;

typedef short sint16;
typedef unsigned short uint16;

typedef int sint32;
typedef unsigned int uint32;

/******************************************************/

#if 0
#define DEBUGP(format, args...) printf(format, ## args)
#else
#define DEBUGP(format, args...)
#endif

#define MAX_PKTSIZE		1512

struct transaction
{
	int length;
	unsigned char data[MAX_PKTSIZE];
};

enum {
	RUNONE = 1,	/* Manually Connect*/
	PERSIST,		/* Always On */
	DEMAND		/* Dial on Demand */
};

/* This structure holds all information necessary to connect/disconnect */
struct session
{
	/* Control paramters */
	char username[64];
	char password[64];
	char authserver[128];
	char authdomain[128];
	char localaddress[64];
	char devname[16]; /* For Router, it is WAN Ethernet interface name */
	uint16 authport;
	uint16 localport;
	int minheartbeat;
	int maxheartbeat;

	int linkmode; /* demand /runone/persist */
	int idletime;

	/* Internal data */
	uint32 sessionid;
	uint16 listenport;
	struct sockaddr_in authhost;
	char osname[80];
	char osrelease[80];
	int listensock;
	struct sockaddr_in localaddr;
	struct sockaddr_in localipaddress;

	uint16 protocol;
	uint16 loginserviceport;
	char loginserverhost[128];
	uint16 hashmethod;
	char nonce[17];
	uint16 retcode;
	uint16 logoutport;
	uint16 statusport;
	char tsmlist[512];
	char tsmlist_s[20][512];
	struct sockaddr_in tsmlist_in[20];
	int tsmcount;
	char resptext[512];
	uint32 timestamp;

	int lastheartbeat;
	int recenthb;
	uint32 sequence;
	struct sockaddr_in fromaddr;
};

/*************************************** main.c ***************************************/
extern int uptime(void);
extern inline int is_dotted_ip(char *addr);
extern void show_bpa_status(int up);


/*************************************** bpalogin.c ***************************************/
extern int mainloop(struct session *s);


/*************************************** md5.c ***************************************/
extern void make_credentials(unsigned char *credentials, struct session *s, uint16 msg, uint32 extra);


/*************************************** demand.c ***************************************/
extern void bpa_demand(struct session *s);
extern void set_bpa_mode(struct session *s, int shutdown, int auth);
extern unsigned long bpa_idle_time(struct session *s);

/*************************************** protocol.c ***************************************/
extern int bpa_login(struct session *s);
extern int bpa_logout(uint16 reason, struct session *s);
extern int handle_heartbeats(struct session *s);


/*************************************** transaction.c ***************************************/
extern void add_field_string(struct transaction *t, uint16 fn, char *p);
extern void add_field_data(struct transaction *t, uint16 fn, char *p, int c);
extern void add_field_int2(struct transaction *t, uint16 fn, uint16 v);
extern void add_field_int4(struct transaction *t, uint16 fn, uint32 v);

extern int extract_int2_value(struct transaction *t, uint16 parm, uint16 *v);
extern int extract_int4_value(struct transaction *t, uint16 parm, uint32 *v);
extern int extract_string_value(struct transaction *t, uint16 parm, char *v);

extern void start_transaction(struct transaction *t, uint16 msgtype, uint32 sessionid);
extern void send_transaction(int socket, struct transaction *t);
extern void send_udp_transaction(struct session *s, struct transaction *t);
extern uint16 receive_transaction(int socket, struct transaction *t);
extern uint16 receive_udp_transaction(struct session *s, int socket, struct transaction *t, 
				struct sockaddr_in *addr);


