/*
  * Implement the BigPond Dial-on-Demand function by referring to ppp-2.4.3.
  * Use a `ETH_P_ALL` socket to capture all packets sent and received from WAN
  * side, these packets are full ethernet frames, and we only check these sent from
  * WAN interface.
  *                                  HY.W (K) 2008 Delta Networks Inc.
  */

#include "bpalogin.h"

/* Open the fd for getting demand-dial packets */
static int open_bpa_loopback(char *devname, uint32 *addr)
{
	int fd;
	struct ifreq ifr;
	struct sockaddr me;
	struct sockaddr_in *our_ip;

	/************************************************************/
	fd = socket(AF_INET, SOCK_RAW, IPPROTO_RAW);
	if (fd < 0)
		return -1;

	memset(&ifr, 0, sizeof(struct ifreq));
	ifr.ifr_addr.sa_family = AF_INET;
	strcpy(ifr.ifr_name, devname);
	if (ioctl(fd, SIOCGIFADDR, &ifr) != 0) {
		printf("BPA: can't get %s's IP!\n", devname);
		close(fd);
		return -1;
	}
	our_ip = (struct sockaddr_in *) &ifr.ifr_addr;
	*addr = our_ip->sin_addr.s_addr;
	DEBUGP("%s (our ip) = %s\n", ifr.ifr_name, inet_ntoa(our_ip->sin_addr));
	close(fd);

	/************************************************************/
	fd = socket(PF_PACKET, SOCK_PACKET, htons(ETH_P_ALL));
	if (fd < 0) {
		printf("BPA: PF_PACKET socket call failed, %s\n", strerror(errno));
		return -1;
	}

	me.sa_family = PF_PACKET;
	strncpy(me.sa_data, devname, 14);
	if (bind(fd, &me, sizeof(me)) < 0) {
		printf("BPA: AF_PACKET bind call failed, %s\n", strerror(errno));
		close(fd);
		return -1;
	}

	return fd;
}

/*
  * Scan a packet to decide whether it is an "active" packet,
  * that is, whether it is worth bringing up the link for.
  *
  * The packet is: iphdr + udp/tcp/icmp ...
  */

#define IP_ICMP	1
#define IP_IGMP	2
#define IP_TCP	6
#define IP_UDP	17
#define IPHDR_OFFSET		0x1FFF

#define ETH_PROTO_IP	0x0800

#define NIPQUAD(addr) \
	((unsigned char *)&addr)[0], \
	((unsigned char *)&addr)[1], \
	((unsigned char *)&addr)[2], \
	((unsigned char *)&addr)[3]

static int active_packet(unsigned char *p, int len, uint32 ourip)
{
	uint8 proto;
	uint32 iphdr; /* ip header length */
	uint32 saddr, daddr;

	if (len < (14 + 20)) {
		DEBUGP("Packet size is too small!\n");
		return 0;
	}

	/*************************************************
		struct ethhdr {
		unsigned char	h_dest[6];		: 00
		unsigned char	h_source[6];		: 06
		__be16		h_proto;			: 12
		} __attribute__((packed));
	*************************************************/
	uint16 ethproto;

	/* Skip the non-IP packets */
	ethproto = *((uint16 *)(p + 12));
	if (ethproto != htons(ETH_PROTO_IP))
		return 0;

	/* Skip the ethernet frame header, point to IP header. */
	p += 14;

	/************************************************* 
		struct iphdr {
		__u8	ihl:4,		: 00 
			version:4;
		__u8	tos;		: 01
		__u16	tot_len;	: 02
		__u16	id;		: 04
		__u16	frag_off;	: 06
		__u8	ttl;		: 08
		__u8	protocol;	: 09
		__u16	check;	: 10
		__u32	saddr;	: 12
		__u32	daddr;	: 16
		};
	*************************************************/
	saddr = * ((uint32 *) (p + 12));

	/* OnlY Packets sent from WAN can bring up the link. */
	if (saddr != ourip)
		return 0;

	uint16 frag_off;

	/* Skip the fragment packets ... */
	frag_off = *((uint16 *)(p + 6));
	if (frag_off & htons(IPHDR_OFFSET))
		return 0;

	/* 
	  * [NETGEAR SPEC V1.6] 5.5 PPP dial on demand: 
	  * The idle time-out & dial-on demand should't treat control packets as traffic,
	  * it includes BPA heartbeat, PPP, PPTP, DDNS query, NTP, TMSS, RIP, and FW
	  * update detection. 
	  */
	proto = p[9];
	if (proto == IP_IGMP) {
		DEBUGP("BigPond: Sending IGMP packet.\n");
		return 0;
	}

	iphdr = (p[0] & 0xF) << 2;
	daddr = * ((uint32 *) (p + 16));

	if (proto == IP_ICMP) {
		printf("BigPond: open ICMP %u.%u.%u.%u -> %u.%u.%u.%u\n",
						NIPQUAD(saddr), NIPQUAD(daddr));
		return 1;
	}
	else if (proto == IP_TCP) {
	/*************************************************
		struct tcphdr {
		__u16	source;		: 00
		__u16	dest;		: 02
		__u32	seq;
		__u32	ack_seq;
		__u16	res1:4, doff:4, 	fin:1, syn:1, rst:1,
				psh:1, ack:1, urg:1, ece:1, cwr:1;	
		__u16	window;
		__u16	check;
		__u16	urg_ptr;
		};
	*************************************************/
		uint8 *tcphdr;
		uint16 sport, dport;

		tcphdr = &p[iphdr];
		/* Assume IP size is 20, not consider data align access fault. */
		dport = *((uint16 *) (tcphdr + 2));
		if (dport == htons(5050)) /* ?_? */
			return 0;

		sport = *((uint16 *) (tcphdr + 0));
		printf("BigPond: open TCP %u.%u.%u.%u:%d -> %u.%u.%u.%u:%d\n",
						NIPQUAD(saddr), ntohs(sport),
						NIPQUAD(daddr), ntohs(dport));
		return 1;
	}
	else if (proto != IP_UDP) {
		printf("BigPond: open %d %u.%u.%u.%u -> %u.%u.%u.%u\n",
						proto, NIPQUAD(saddr), NIPQUAD(daddr));
		return 1;
	}

	/*************************************************	
		struct udphdr {
		__u16	source;	: 00
		__u16	dest;	: 02
		__u16	len;		: 04
		__u16	check;	: 06
		};
	*************************************************/
	uint8 *udphdr;
	uint16 sport, dport;

	udphdr = &p[iphdr];
	dport = *((uint16 *) (udphdr + 2));
	dport = ntohs(dport);

	if (dport == 123) {
		DEBUGP("BigPond: Sending NTP packet.\n");
		return 0;
	}
	else if (dport == 53) {
		/* 
		  * Verify the DNS query packet size ?_?
 		  * DNS packet has 12 fixed octet header, then the domain name and etc ...
 		  * the domain is formatted as bellow:
 		  * 	gemini.tuc.noao.edu --> 6gemini3tuc4noao3edu0
		  */
		char *dns = (char *)udphdr + (8 + 13);
		if (strncmp(dns, "time-", 5) == 0) {
			DEBUGP("BigPond: Sending a time-*.netgear.com DNS query.\n");
			return 0;
		}
	}
	else if (dport == 520) {
		DEBUGP("BigPond: Sending a RIP packet.\n");
		return 0;
	}
	else if (dport == 67) {
		DEBUGP("BigPond: Sending a DHCP packet.\n");
		return 0;
	}

	sport = *((uint16 *) (udphdr + 0));
	printf("BigPond: open UDP %u.%u.%u.%u:%d -> %u.%u.%u.%u:%d\n",
					NIPQUAD(saddr), ntohs(sport),
					NIPQUAD(daddr), dport);
	return 1;
}

void bpa_demand(struct session *s)
{
	int r;
	int fd_loop;
	int buf_size;
	uint32 ourip;
	fd_set readable;
	unsigned char *buf, packet[1520];

	fd_loop = open_bpa_loopback(s->devname, &ourip);
	if (fd_loop == -1) {
		printf("BPA: can't open the loop fd for dial-on-demand!\n");
		return;
	}

	/* Ethernet Header size is 14, for data align access .... */
	buf = &packet[2]; buf_size = sizeof(packet) - 2;
	DEBUGP("BigPond is waiting traffic to bring up the link ...\n");
	for (;;) {
		FD_ZERO(&readable);
		FD_SET(fd_loop, &readable);

		r = select(fd_loop + 1, &readable, NULL, NULL, NULL);
		if (r < 1)
			continue;

		r = recvfrom(fd_loop, buf, buf_size, 0, NULL, NULL);
		if (active_packet(buf, r, ourip)) {
			close(fd_loop);
			return;
		}
	}
}

/********************************************************************************/
#define IFF_BPA_MODE	0x20	/* `1`: indicate 'BigPond' mode */
#define IFF_BPA_AUTH		0x40	/* `1`: can xmit/recv traffic; '0': limit traffic passthough */

#define SIOCEIDLETIME	0x89F1 /* get idle time of eth1 */
#define SIODNIBPAMODE	0x89F2  /* set the bpa mode */

void set_bpa_mode(struct session *s, int shutdown, int auth)
{
	int fd;
	struct ifreq ifr;
	unsigned short flags;

	fd = socket(AF_INET, SOCK_DGRAM, 0);
	if (fd < 0) {
		printf("BPA mode: socket failed, %s\n", strerror(errno));
		return;
	}

	memset(&ifr, 0, sizeof(struct ifreq));
	strcpy(ifr.ifr_name, s->devname);
	ifr.ifr_data = (char *)&flags;

	if (shutdown == 0)
		flags = (IFF_BPA_MODE) |(auth ? IFF_BPA_AUTH : 0);
	else
		flags = 0;

	if (ioctl(fd, SIODNIBPAMODE, &ifr) < 0 )
		printf("BPA mode: can't set BigPond mode on %s, %s\n",
					s->devname, strerror(errno));

	close(fd);
}

unsigned long bpa_idle_time(struct session *s)
{
	int fd;
	struct ifreq ifr;
	unsigned long idletime;

	fd = socket(AF_INET, SOCK_DGRAM, 0);
	if (fd < 0) {
		printf("BPA Idle: socket failed, %s\n", strerror(errno));
		return 0xFFFFFFFF; /* Never Idle */
	}

	memset(&ifr, 0, sizeof(struct ifreq));
	strcpy(ifr.ifr_name, s->devname);
	ifr.ifr_data = (char *)&idletime;
	if (ioctl(fd, SIOCEIDLETIME, &ifr) < 0 ) {
		static int warning;

		if (warning++ < 3)
			printf("BPA Idle: can't get %s idle time, %s\n", s->devname,
					strerror(errno));
		idletime = 0xFFFFFFFF;
	}

	close(fd);

	return idletime;
}

